from functools import partial
import torch
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence


def dict_collate_fn(batch, pad_value):
    """
    Custom collate function to handle dictionaries in a batch.
    """
    assert isinstance(batch[0], dict), "Batch must be a dictionary."
    keys = batch[0].keys()
    batch = [d.values() for d in batch]

    output = [
        (
            pad_sequence(d, batch_first=True, padding_value=pad_value)
            if torch.is_tensor(d[0])
            else list(d)
        )
        for d in zip(*batch)
    ]
    return dict(zip(keys, output))


class DictDataLoader(DataLoader):
    """
    A DataLoader that returns a dictionary of data.
    """

    def __init__(self, *args, **kwargs):
        pad_value = kwargs.pop("pad_value", -1)
        collate_fn = kwargs.pop(
            "collate_fn", partial(dict_collate_fn, pad_value=pad_value)
        )
        super().__init__(collate_fn=collate_fn, *args, **kwargs)



import time
import torch
import pytorch_lightning as pl


class SaveCheckpoint(pl.callbacks.ModelCheckpoint):
    def __init__(self, **kwargs):
        """
        Initializes the callback with the given configuration.
        Args:
            kwargs (dict): A configuration dictionary containing the following keys:
                - dirpath (str, optional): Directory path where checkpoints will be saved.
                  Defaults to "checkpoints/".
                - filename (str, optional): Filename format for the checkpoints.
                  Defaults to "step-{step}".
                - save_top_k (int, optional): Number of best models to save.
                  Defaults to -1 (save all checkpoints).
                - every_n_train_steps (int, optional): Frequency (in training steps)
                  at which checkpoints are saved. Defaults to 512.
                - save_weights_only (bool, optional): Whether to save only model weights
                  instead of the full model. Defaults to False.
                - **kwargs: Additional keyword arguments passed to the parent class initializer.
        """

        dirpath = kwargs.pop("dirpath", "checkpoints/")
        filename = kwargs.pop("filename", "{step:05d}")
        save_top_k = kwargs.pop("save_top_k", -1)
        every_n_train_steps = kwargs.pop("every_n_train_steps", 512)
        save_weights_only = kwargs.pop("save_weights_only", False)
        super().__init__(
            dirpath=dirpath,
            filename=filename,
            save_top_k=save_top_k,
            every_n_train_steps=every_n_train_steps,
            save_weights_only=save_weights_only,
            save_last=True,
            monitor="step",
            mode="max",
            **kwargs,
        )
        print(
            "Save checkpoint strategy initialized with the following parameters:"
            f"\n- dirpath: {dirpath}"
            f"\n- filename: {filename}"
            f"\n- save_top_k: {save_top_k}"
            f"\n- every_n_train_steps: {every_n_train_steps}"
            f"\n- save_weights_only: {save_weights_only}"
        )


class LogLearningRate(pl.Callback):
    def __init__(self):
        super().__init__()

    def on_train_batch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule, outputs, batch, batch_idx):
        # Get the learning rate from the optimizer
        for param_group in trainer.optimizers[0].param_groups:
            lr = param_group["lr"]
            break
        pl_module.log("trainer/lr", lr, on_step=True, logger=True, sync_dist=True)


class LogGradient(pl.Callback):
    def __init__(self, should_stop: bool = False):
        super().__init__()
        self.should_stop = should_stop

    def on_after_backward(self, trainer, pl_module):
        total_norm = pl_module.total_norm
        if total_norm is None:
            return
        pl_module.log("trainer/norm", total_norm, on_step=True, logger=True, sync_dist=True)
        if torch.isinf(total_norm) or torch.isnan(total_norm):
            print(f"Infinite/NaN gradient norm @ {trainer.current_epoch} epoch.")
            trainer.save_checkpoint(
                f"checkpoints/inf_nan_gradient_epoch_{trainer.current_epoch}.ckpt",
                weights_only=True,
            )
            trainer.should_stop = self.should_stop


class LogETL(pl.Callback):
    def on_fit_start(self, trainer, pl_module):
        self.start_time = time.time()

    def on_train_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        if trainer.max_epochs==-1:
            return
        elapsed_time = time.time() - self.start_time
        elapsed_epochs = trainer.current_epoch - pl_module.start_epoch
        if elapsed_epochs < 1:
            pl_module.start_epoch = trainer.current_epoch + 1
            elapsed_epochs = 1
        remaining_time = (elapsed_time / elapsed_epochs) * (
            trainer.max_epochs - trainer.current_epoch
        )
        pl_module.log(
            "trainer/ETL (min)", remaining_time / 60, on_epoch=True, logger=True, sync_dist=True
        )

    def on_train_batch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule, outputs, batch, batch_idx):
        if trainer.max_steps==-1:
            return
        elapsed_time = time.time() - self.start_time
        elapsed_steps = trainer.global_step - pl_module.start_step + 1
        if elapsed_steps < 1:
            pl_module.start_step = trainer.global_step + 1
            elapsed_steps = 1
        remaining_time = (elapsed_time / elapsed_steps) * (
            trainer.max_steps - trainer.global_step
        )
        pl_module.log(
            "trainer/ETL (min)", remaining_time / 60, on_step=True, logger=True, sync_dist=True
        )


import torch

class EMA:
    def __init__(self, parameters, update_after_step=100, inv_gamma=1.0, power=0.75, min_value=0.0, max_value=0.9999):
        self.shadow = {}
        self.backup = {}
        self.update_after_step = update_after_step
        self.inv_gamma = inv_gamma
        self.power = power
        self.min_value = min_value
        self.max_value = max_value
        self.step_count = 0

        for name, param in parameters:
            if param.requires_grad:
                self.shadow[name] = param.data.clone()

    def _get_decay(self):
        step = max(0, self.step_count - self.update_after_step)
        if step <= 0:
            return 0.0
        decay = 1.0 - (1.0 + step / self.inv_gamma) ** -self.power
        return max(self.min_value, min(decay, self.max_value))

    @torch.no_grad()
    def update(self, named_params):
        decay = self._get_decay()
        self.decay = decay
        for name, param in named_params:
            if name not in self.shadow:
                self.shadow[name] = param.data.clone().detach().to(param.device)
            else:
                self.shadow[name] = (
                    self.shadow[name].to(param.device)
                    .mul(decay)
                    .add(param.data, alpha=1.0 - decay)
                )
        self.step_count += 1

    def apply_to(self, model):
        self.backup = {}
        for name, param in model.named_parameters():
            if name in self.shadow:
                self.backup[name] = param.data.clone()
                param.data.copy_(self.shadow[name])

    def restore(self, model):
        for name, param in model.named_parameters():
            if name in self.backup:
                param.data.copy_(self.backup[name])
        self.backup = {}

    def state_dict(self):
        return {
            "shadow": self.shadow,
            "step_count": self.step_count,
            "update_after_step": self.update_after_step,
            "inv_gamma": self.inv_gamma,
            "power": self.power,
            "min_value": self.min_value,
            "max_value": self.max_value
        }

    def load_state_dict(self, state_dict):
        self.shadow = state_dict["shadow"]
        self.step_count = state_dict["step_count"]
        self.update_after_step = state_dict["update_after_step"]
        self.inv_gamma = state_dict["inv_gamma"]
        self.power = state_dict["power"]
        self.min_value = state_dict["min_value"]
        self.max_value = state_dict["max_value"]